# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment as neutron_env

from networking_bagpipe.tests.fullstack.resources.bgpvpn \
    import client as bgpvpn_client
from networking_bagpipe.tests.fullstack.resources.common \
    import environment as common_env

SUBNET_CIDR1 = '10.0.0.0/24'
SUBNET_CIDR2 = '20.0.0.0/24'
SUBNET_CIDR3 = '30.0.0.0/24'


class BaGPipeBaseFullStackTestCase(base.BaseFullStackTestCase):

    evpn_driver = 'dummy'
    ipvpn_driver = 'dummy'

    compute_node_count = 3
    port_per_compute_per_net = 2

    def setUp(self):
        host_descriptions = [
            neutron_env.HostDescription(of_interface=self.of_interface,
                                        l2_agent_type=self.l2_agent_type)
            for _ in range(self.compute_node_count)
        ]
        env = common_env.BaGPipeEnvironment(
            common_env.BaGPipeEnvironmentDescription(
                bagpipe_ml2=self.bagpipe_ml2,
                evpn_driver=self.evpn_driver,
                bgpvpn=self.bgpvpn,
                ipvpn_driver=self.ipvpn_driver,
                ipvpn_encap=self.ipvpn_encap,
                mech_drivers=self.mech_drivers,
                service_plugins=self.service_plugins
            ),
            host_descriptions)
        super(BaGPipeBaseFullStackTestCase, self).setUp(env)

        if self.bgpvpn:
            self.safe_client = self.useFixture(
                bgpvpn_client.BGPVPNClientFixture(self.client))

    def _create_net_subnet_bgpvpn_assoc(self, tenant_uuid, subnet_cidr,
                                        bgpvpn_id=None):
        network = self.safe_client.create_network(tenant_uuid)
        subnet = self.safe_client.create_subnet(
            tenant_uuid, network['id'], subnet_cidr)

        if bgpvpn_id:
            self.safe_client.create_network_association(tenant_uuid,
                                                        bgpvpn_id,
                                                        network['id'])

        return (network['id'], subnet['id'])

    def _create_router_bgpvpn_assoc(self, tenant_uuid, subnet_ids,
                                    bgpvpn_id=None):
        router = self.safe_client.create_router()

        for subnet_id in subnet_ids:
            self.safe_client.add_router_interface(router['id'], subnet_id)

        self.safe_client.create_router_association(tenant_uuid,
                                                   bgpvpn_id,
                                                   router['id'])



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from networking_bagpipe.agent.bagpipe_ml2 import agent_extension as ml2_agt_ext
from networking_bagpipe.agent.bgpvpn import agent_extension as bgpvpn_agt_ext
from networking_bagpipe.bagpipe_bgp.api import api
from networking_bagpipe.bagpipe_bgp.common import config
from networking_bagpipe.bagpipe_bgp.common import run_command
from networking_bagpipe.bagpipe_bgp import constants
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers
from networking_bagpipe.bagpipe_bgp.vpn.evpn import linux_vxlan
from networking_bagpipe.bagpipe_bgp.vpn.ipvpn import mpls_linux_dataplane
from networking_bagpipe.bagpipe_bgp.vpn.ipvpn import mpls_ovs_dataplane
from networking_bagpipe.driver import sfc


# NOTE(amotoki): oslo.config suggests to use lower case as group name.
# If a group name is registered with upper case names,
# oslo.config looks up both upper case and lower case versions of names
# in configuration files, so using lower case in sample files is safe enough.


def list_bagpipe_bgp_agent_opts():
    return [
        ('bagpipe_ml2_extension', ml2_agt_ext.opts),
        ('bagpipe', bgpvpn_agt_ext.bagpipe_bgpvpn_opts),
    ]


def list_api_opts():
    return [
        ('api', api.common_opts),
    ]


def list_bgp_common_opts():
    return [
        ('bgp', config.bgp_opts),
    ]


def list_run_command_opts():
    return [
        ('common', run_command.common_opts),
    ]


def list_dataplane_driver_ipvpn_opts():
    return [
        (constants.config_group(constants.IPVPN).lower(),
         dataplane_drivers.dataplane_common_opts),
    ]


def list_dataplane_driver_evpn_opts():
    return [
        (constants.config_group(constants.EVPN).lower(),
         dataplane_drivers.dataplane_common_opts),
    ]


def list_dataplane_driver_evpn_linux_vxlan_opts():
    return [
        (constants.config_group(constants.EVPN).lower(),
         linux_vxlan.LinuxVXLANDataplaneDriver.driver_opts),
    ]


def list_dataplane_driver_ipvpn_mpls_linux_opts():
    return [
        (constants.config_group(constants.IPVPN).lower(),
         mpls_linux_dataplane.MPLSLinuxDataplaneDriver.driver_opts),
    ]


def list_dataplane_driver_ipvpn_mpls_ovs_opts():
    return [
        (constants.config_group(constants.IPVPN).lower(),
         mpls_ovs_dataplane.MPLSOVSDataplaneDriver.driver_opts),
    ]


def list_neutron_sfc_opts():
    return [
        ('sfc_bagpipe', sfc.sfc_bagpipe_opts)
    ]



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
import sqlalchemy as sa

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lib import context as n_context
from neutron_lib.db import model_base

from neutron.db import common_db_mixin

LOG = logging.getLogger(__name__)


class BaGPipePpgRTAssoc(model_base.BASEV2, model_base.HasId):
    '''Mapping from Port Pair Group to the NN part of an AS:NN route target'''

    __tablename__ = 'sfc_bagpipe_ppg_rtnn_associations'
    ppg_id = sa.Column(sa.String(36), primary_key=True)
    rtnn = sa.Column(sa.Integer, unique=True, nullable=False)
    is_redirect = sa.Column(sa.Boolean(), nullable=False)
    reverse = sa.Column(sa.Boolean(), nullable=False)


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance


@singleton
class RTAllocator(object):
    def __init__(self, config):
        self.config = config
        self.session = n_context.get_admin_context().session

    def _get_rt_from_rtnn(self, rtnn):
        return ':'.join([str(self.config.as_number), str(rtnn)])

    @log_helpers.log_method_call
    def allocate_rt(self, ppg_id, is_redirect=False, reverse=False):
        query = self.session.query(
            BaGPipePpgRTAssoc).order_by(
            BaGPipePpgRTAssoc.rtnn)

        allocated_rtnns = {obj.rtnn for obj in query.all()}

        # Find first one available in range
        start, end = int(self.config.rtnn[0]), int(self.config.rtnn[1]) + 1
        for rtnn in six.moves.range(start, end):
            if rtnn not in allocated_rtnns:
                with self.session.begin(subtransactions=True):
                    ppg_rtnn = BaGPipePpgRTAssoc(ppg_id=ppg_id,
                                                 rtnn=rtnn,
                                                 is_redirect=is_redirect,
                                                 reverse=reverse)
                    self.session.add(ppg_rtnn)
                return self._get_rt_from_rtnn(rtnn)

        LOG.error("Can't allocate route target, all range in use")
        return None

    @log_helpers.log_method_call
    def get_rts_by_ppg(self, ppg_id):

        ppg_rts = self.session.query(
            BaGPipePpgRTAssoc).filter_by(ppg_id=ppg_id).all()
        if not ppg_rts:
            return None

        return [ppg_rt.rtnn for ppg_rt in ppg_rts]

    @log_helpers.log_method_call
    def get_redirect_rt_by_ppg(self, ppg_id, reverse=False):
        ppg_redirect_rt = self.session.query(
            BaGPipePpgRTAssoc).filter_by(ppg_id=ppg_id,
                                         is_redirect=True,
                                         reverse=reverse).one()

        return self._get_rt_from_rtnn(ppg_redirect_rt.rtnn)

    @log_helpers.log_method_call
    def release_rt(self, rtnn):
        with self.session.begin(subtransactions=True):
            ppg_rtnn = self.session.query(
                BaGPipePpgRTAssoc).filter_by(rtnn=rtnn).first()

            if ppg_rtnn:
                self.session.delete(ppg_rtnn)
            else:
                LOG.warning("Can't release route target %s, not used", rtnn)


class BaGPipeChainHop(model_base.BASEV2, model_base.HasId,
                      model_base.HasProject):
    __tablename__ = 'sfc_bagpipe_chain_hops'
    ingress_ppg = sa.Column(sa.String(36),
                            nullable=True)
    egress_ppg = sa.Column(sa.String(36),
                           nullable=True)
    ingress_network = sa.Column(sa.String(36),
                                sa.ForeignKey('networks.id'),
                                nullable=True)
    egress_network = sa.Column(sa.String(36),
                               sa.ForeignKey('networks.id'),
                               nullable=True)
    rts = sa.Column(sa.String(255))
    ingress_gw = sa.Column(sa.String(64))
    egress_gw = sa.Column(sa.String(64))
    readv_from_rts = sa.Column(sa.String(255))
    readv_to_rt = sa.Column(sa.String(255))
    attract_to_rt = sa.Column(sa.String(255))
    redirect_rts = sa.Column(sa.String(255))
    classifiers = sa.Column(sa.String(255))
    reverse_hop = sa.Column(sa.Boolean(), nullable=False)
    portchain_id = sa.Column(sa.String(36),
                             nullable=False,
                             primary_key=True)


class BaGPipeSfcDriverDB(common_db_mixin.CommonDbMixin):

    def initialize(self):
        self.admin_context = n_context.get_admin_context()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import random

import netaddr
from oslo_utils import uuidutils

from networking_bagpipe.objects import bgpvpn as bgpvpn_obj

from neutron.common import utils
from neutron.objects import network
from neutron.objects import ports
from neutron.objects import router
from neutron.objects import subnet

from neutron.tests.unit.objects import test_base
from neutron.tests.unit import testlib_api

from neutron_lib.api.definitions import bgpvpn as bgpvpn_api
from neutron_lib.api.definitions import bgpvpn_routes_control as bgpvpn_rc_api
from neutron_lib import constants
from neutron_lib import context


test_base.FIELD_TYPE_VALUE_GENERATOR_MAP[bgpvpn_obj.BGPVPNTypeField] = (
    lambda: random.choice(bgpvpn_api.BGPVPN_TYPES)
)

test_base.FIELD_TYPE_VALUE_GENERATOR_MAP[
    bgpvpn_obj.BGPVPNPortAssociationRouteTypeField] = (
        # do not generate bgpvpn type routes for now:
        lambda: bgpvpn_rc_api.PREFIX_TYPE)


CIDR = "10.10.0.0/16"
GW_IP = "10.10.0.1"
GW_MAC = "ba:ad:00:00:ca:fe"
TEST_RT = "64512:42"


def _subnet_dict(gw_mac=None):
    return {
        'id': mock.ANY,
        'ip_version': 4,
        'gateway_mac': gw_mac,
        'cidr': utils.AuthenticIPNetwork(CIDR),
        'gateway_ip': netaddr.IPAddress(GW_IP)
    }


class _BPGVPNObjectsTestCommon(object):

    def _create_test_bgpvpn(self):
        bgpvpn = bgpvpn_obj.BGPVPN(self.context,
                                   route_targets=[TEST_RT],
                                   name='test-bgpvpn-U',
                                   type='l3')
        bgpvpn.create()
        return bgpvpn

    def _create_test_bgpvpn_id(self):
        return self._create_test_bgpvpn().id

    def _make_subnet(self, network_id):
        _subnet = subnet.Subnet(self.context,
                                network_id=network_id,
                                ip_version=4,
                                cidr=netaddr.IPNetwork(CIDR),
                                gateway_ip=GW_IP)
        _subnet.create()
        return _subnet

    def _connect_router_network(self, router_id, network_id,
                                subnet_id=None, gw_network=False):
        port = ports.Port(self.context,
                          network_id=network_id,
                          mac_address=netaddr.EUI(
                              GW_MAC,
                              dialect=netaddr.mac_unix_expanded),
                          device_id='test_device_id',
                          device_owner=constants.DEVICE_OWNER_ROUTER_INTF,
                          status="DUMMY_STATUS",
                          admin_state_up=True)
        if gw_network:
            port.device_owner = constants.DEVICE_OWNER_ROUTER_GW

        port.create()

        if subnet_id:
            allocation = ports.IPAllocation(
                self.context,
                port_id=port.id,
                subnet_id=subnet_id,
                network_id=network_id,
                ip_address=netaddr.IPNetwork(GW_IP))
            allocation.create()

            port.fixed_ips = [allocation]
            port.update()

        router_if = router.RouterPort(self.context,
                                      router_id=router_id,
                                      port_id=port.id)
        router_if.create()


class BGPVPNTest(test_base.BaseDbObjectTestCase,
                 testlib_api.SqlTestCase,
                 _BPGVPNObjectsTestCommon):

    _test_class = bgpvpn_obj.BGPVPN

    def test_get_objects_supports_extra_filtername(self):
        self.skipTest("no support for extra filtername")

    def test_get_object(self):
        bgpvpn = self._create_test_bgpvpn()
        bgpvpn_bis = bgpvpn_obj.BGPVPN.get_object(self.context,
                                                  id=bgpvpn.id)
        self.assertEqual(bgpvpn, bgpvpn_bis)


class BGPVPNNetAssociationTest(test_base.BaseDbObjectTestCase,
                               testlib_api.SqlTestCase,
                               _BPGVPNObjectsTestCommon):

    _test_class = bgpvpn_obj.BGPVPNNetAssociation

    def setUp(self):
        test_base.BaseDbObjectTestCase.setUp(self)
        self.network_id = self._create_test_network_id()
        self.update_obj_fields(
            {'network_id': self.network_id,
             'bgpvpn_id': self._create_test_bgpvpn_id})

        self.subnet = self._make_subnet(self.network_id)

    def test_get_objects_queries_constant(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_all_subnets(self):
        for db_obj in self.objs:
            self.assertItemsEqual(db_obj.all_subnets(self.network_id),
                                  [_subnet_dict()])

    def test_subnets(self):
        for obj in self.objs:
            obj.create()
            self.assertItemsEqual(obj.subnets, [_subnet_dict()])

        # plug a router
        _router = router.Router(self.context)
        _router.create()

        self._connect_router_network(_router.id,
                                     self.network_id,
                                     self.subnet.id)

        # check .subnets in associations, after refreshing
        for obj in self.objs:
            refreshed_obj = bgpvpn_obj.BGPVPNNetAssociation.get_object(
                self.context,
                id=obj.id)
            self.assertItemsEqual(refreshed_obj.subnets,
                                  [_subnet_dict(GW_MAC)])


class BGPVPNRouterAssociationTest(test_base.BaseDbObjectTestCase,
                                  testlib_api.SqlTestCase,
                                  _BPGVPNObjectsTestCommon):

    _test_class = bgpvpn_obj.BGPVPNRouterAssociation

    def setUp(self):
        test_base.BaseDbObjectTestCase.setUp(self)
        self.router_id = self._create_test_router_id()
        self.update_obj_fields(
            {'router_id': self.router_id,
             'bgpvpn_id': self._create_test_bgpvpn_id,
             'routes': []})
        self.context = context.get_admin_context()

    def test_get_objects_queries_constant(self):
        self.skipTest("test not passing yet, remains to be investigated why")

    def test_all_subnets(self):
        for obj in self.objs:
            obj.create()

        network_id = self._create_test_network_id()
        subnet_ = self._make_subnet(network_id)

        # initially the network is not connected to the router
        for obj in self.objs:
            self.assertItemsEqual(obj.all_subnets(network_id), [])

        self._connect_router_network(self.router_id,
                                     network_id,
                                     subnet_.id)

        # connect a gateway network
        gw_network_id = self._create_test_network_id()
        self._connect_router_network(self.router_id,
                                     gw_network_id,
                                     gw_network=True)

        # check .subnets in associations, after refreshing
        # (except gateway network that should not be present)
        for obj in self.objs:
            refreshed_obj = bgpvpn_obj.BGPVPNRouterAssociation.get_object(
                self.context,
                id=obj.id)
            self.assertItemsEqual(refreshed_obj.all_subnets(network_id),
                                  [_subnet_dict(GW_MAC)])
            self.assertItemsEqual(refreshed_obj.all_subnets("dummy-uuid"),
                                  [])

    def test_get_objects_from_network_id(self):
        router_ = router.Router(self.context)
        router_.create()

        self.project = uuidutils.generate_uuid()

        # put a network behind a router
        network_ = network.Network(self.context)
        network_.create()

        subnet_ = self._make_subnet(network_.id)

        self._connect_router_network(router_.id,
                                     network_.id)

        bgpvpn_ = self._create_test_bgpvpn()

        router_assoc_ = bgpvpn_obj.BGPVPNRouterAssociation(
            self.context,
            project_id=self.project,
            router_id=router_.id,
            bgpvpn_id=bgpvpn_.id)
        router_assoc_.create()

        # unrelated router and BGPVPN
        router_2 = router.Router(self.context)
        router_2.create()
        router_assoc_2 = bgpvpn_obj.BGPVPNRouterAssociation(
            self.context,
            project_id=self.project,
            router_id=router_2.id,
            bgpvpn_id=self._create_test_bgpvpn_id())
        router_assoc_2.create()

        # test get_objects
        get_assocs = bgpvpn_obj.BGPVPNRouterAssociation.get_objects(
            self.context,
            network_id=network_.id)

        self.assertEqual(1, len(get_assocs))
        self.assertEqual(get_assocs[0].bgpvpn.id, bgpvpn_.id)
        self.assertIn(
            subnet_.id,
            [s['id'] for s in get_assocs[0].all_subnets(network_.id)])


class BGPVPNPortAssociationTest(test_base.BaseDbObjectTestCase,
                                testlib_api.SqlTestCase,
                                _BPGVPNObjectsTestCommon):

    _test_class = bgpvpn_obj.BGPVPNPortAssociation

    def setUp(self):
        test_base.BaseDbObjectTestCase.setUp(self)
        self.project = uuidutils.generate_uuid()
        self.port_id = self._create_test_port_id()
        self.update_obj_fields(
            {'port_id': self.port_id,
             'bgpvpn_id': self._create_test_bgpvpn_id,
             'routes': {
                 'bgpvpn_id': self._create_test_bgpvpn_id,
                 }})

    def test_get_objects_queries_constant(self):
        self.skipTest("test not passing yet, remains to be investigated why")


class BGPVPNPortAssociationRouteTest(test_base.BaseDbObjectTestCase,
                                     testlib_api.SqlTestCase,
                                     _BPGVPNObjectsTestCommon):

    _test_class = bgpvpn_obj.BGPVPNPortAssociationRoute

    def setUp(self):
        test_base.BaseDbObjectTestCase.setUp(self)
        self.project = uuidutils.generate_uuid()
        self.update_obj_fields(
            {'port_association_id': self._create_test_port_assoc_id,
             'bgpvpn_id': self._create_test_bgpvpn_id})
        self.context = context.get_admin_context()

    def _create_test_port_assoc(self):
        bgpvpn_id = self._create_test_bgpvpn_id()
        port_id = self._create_test_port_id()
        port_assoc = bgpvpn_obj.BGPVPNPortAssociation(self.context,
                                                      project_id=self.project,
                                                      port_id=port_id,
                                                      bgpvpn_id=bgpvpn_id)
        port_assoc.create()
        return port_assoc

    def _create_test_port_assoc_id(self):
        return self._create_test_port_assoc().id

    def test_eq_hash_prefix(self):
        r1 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='prefix',
            prefix=netaddr.IPNetwork('1.2.3.4'))
        r2 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='prefix',
            prefix=netaddr.IPNetwork('1.2.3.4'))
        self.assertTrue(r1 == r2)
        self.assertTrue(hash(r1) == hash(r2))

    def test_eq_hash_bgpvpn(self):
        bgpvpn = self._create_test_bgpvpn()
        r1 = bgpvpn_obj.BGPVPNPortAssociationRoute(type='bgpvpn',
                                                   bgpvpn=bgpvpn)
        r2 = bgpvpn_obj.BGPVPNPortAssociationRoute(type='bgpvpn',
                                                   bgpvpn=bgpvpn)
        self.assertTrue(r1 == r2)
        self.assertTrue(hash(r1) == hash(r2))

    def test_neq_type(self):
        r1 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='bgpvpn',
            bgpvpn_id='12345')
        r2 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='prefix',
            prefix=netaddr.IPNetwork('1.2.3.4'))
        self.assertTrue(r1 != r2)

    def test_neq_prefix(self):
        r1 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='prefix',
            prefix=netaddr.IPNetwork('11.22.33.44'))
        r2 = bgpvpn_obj.BGPVPNPortAssociationRoute(
            type='prefix',
            prefix=netaddr.IPNetwork('1.2.3.4'))
        self.assertTrue(r1 != r2)

    def test_neq_bgpvpn(self):
        bgpvpn1 = self._create_test_bgpvpn()
        bgpvpn2 = self._create_test_bgpvpn()
        r1 = bgpvpn_obj.BGPVPNPortAssociationRoute(type='bgpvpn',
                                                   bgpvpn=bgpvpn1)
        r2 = bgpvpn_obj.BGPVPNPortAssociationRoute(type='bgpvpn',
                                                   bgpvpn=bgpvpn2)
        self.assertTrue(r1 != r2)

    def test_bgpvpn_route_get_object_access_bgpvpn(self):
        route_id = uuidutils.generate_uuid()
        route = bgpvpn_obj.BGPVPNPortAssociationRoute(
            self.context,
            id=route_id,
            port_association_id=self._create_test_port_assoc().id,
            type=bgpvpn_rc_api.BGPVPN_TYPE,
            bgpvpn_id=self._create_test_bgpvpn_id())
        route.create()

        route_again = bgpvpn_obj.BGPVPNPortAssociationRoute.get_object(
            self.context,
            id=route_id
            )

        self.assertEqual([TEST_RT], route_again.bgpvpn.route_targets)



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Defining SFC data-model

Revision ID: d2c2dcb6c2d4
Revises: 6185f1633a3d
Create Date: 2017-03-02 15:59:58.430218

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd2c2dcb6c2d4'
down_revision = 'd4d4d7f03b21'


def upgrade():
    op.create_table(
        'sfc_bagpipe_ppg_rtnn_associations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('ppg_id', sa.String(length=36),  nullable=False),
        sa.Column('rtnn', sa.Integer(), nullable=False),
        sa.Column('is_redirect', sa.Boolean(), nullable=False),
        sa.Column('reverse', sa.Boolean(), nullable=False),

        sa.PrimaryKeyConstraint('id', 'ppg_id'),
        sa.UniqueConstraint('rtnn')
    )

    op.create_table(
        'sfc_bagpipe_chain_hops',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('portchain_id', sa.String(length=36), nullable=False),
        sa.Column('rts', sa.String(length=255), nullable=True),
        sa.Column('ingress_gw', sa.String(length=64), nullable=False),
        sa.Column('egress_gw', sa.String(length=64), nullable=False),
        sa.Column('ingress_ppg', sa.String(length=36), nullable=True),
        sa.Column('egress_ppg', sa.String(length=36), nullable=True),
        sa.Column('ingress_network', sa.String(length=36), nullable=True),
        sa.Column('egress_network', sa.String(length=36), nullable=True),
        sa.Column('readv_from_rts', sa.String(length=255), nullable=True),
        sa.Column('readv_to_rt', sa.String(length=255), nullable=True),
        sa.Column('attract_to_rt', sa.String(length=255), nullable=True),
        sa.Column('redirect_rts', sa.String(length=255), nullable=True),
        sa.Column('classifiers', sa.String(length=255), nullable=True),
        sa.Column('reverse_hop', sa.Boolean(), nullable=False),

        sa.ForeignKeyConstraint(['ingress_network'],
                                ['networks.id']),
        sa.ForeignKeyConstraint(['egress_network'],
                                ['networks.id']),
        sa.PrimaryKeyConstraint('id', 'portchain_id')
    )



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from networking_bagpipe.agent.sfc import agent_extension as bagpipe_agt_ext
from networking_bagpipe.bagpipe_bgp import constants as bbgp_const
from networking_bagpipe.driver import constants as sfc_const
from networking_bagpipe.objects import sfc as sfc_obj
from networking_bagpipe.tests.unit.agent import base

from neutron.api.rpc.callbacks.consumer import registry
from neutron.api.rpc.callbacks import events as rpc_events
from neutron.api.rpc.handlers import resources_rpc

from neutron.plugins.ml2.drivers.linuxbridge.agent.common \
    import constants as lb_agt_constants
from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager

CHAIN_HOP_CLASSIFIER = {"protocol": "tcp",
                        "sourcePrefix": "1.2.3.4/32",
                        "destinationPort": "80",
                        "destinationPrefix": "5.6.7.8/32"}

CHAIN_HOP_REVERSE_CLASSIFIER = {"protocol": "tcp",
                                "sourcePrefix": "5.6.7.8/32",
                                "sourcePort": "80",
                                "destinationPrefix": "1.2.3.4/32"}

CHAIN_HOP_RT1000 = ['SFC_L3:1000']
CHAIN_HOP_EGRESS_TO_RT1002 = 'SFC_L3:1002'
CHAIN_HOP_EGRESS_PARAMS1 = {'readv_from_rts': ['SFC_L3:1001'],
                            'readv_to_rt': CHAIN_HOP_EGRESS_TO_RT1002,
                            'redirect_rts': ['SFC_L3:1003']}

CHAIN_HOP_RT2000 = ['SFC_L3:2000']
CHAIN_HOP_EGRESS_TO_RT2002 = 'SFC_L3:2002'
CHAIN_HOP_EGRESS_PARAMS2 = {'readv_from_rts': ['SFC_L3:2001'],
                            'readv_to_rt': CHAIN_HOP_EGRESS_TO_RT2002,
                            'redirect_rts': ['SFC_L3:2003']}

net_ports = {
    base.NETWORK1['id']: [base.PORT10['id'], base.PORT11['id']],
    base.NETWORK2['id']: [base.PORT20['id'], base.PORT21['id']]
}


class TestSfcAgentExtension(base.BaseTestLinuxBridgeAgentExtension):

    agent_extension_class = bagpipe_agt_ext.BagpipeSfcAgentExtension

    def setUp(self):
        super(TestSfcAgentExtension, self).setUp()
        self.mocked_bulk_rpc = mock.patch.object(
            self.agent_ext._pull_rpc, 'bulk_pull').start()

    @mock.patch.object(registry, 'register')
    @mock.patch.object(resources_rpc, 'ResourcesPushRpcCallback')
    def test_initialize_rpcs(self, rpc_mock, subscribe_mock):
        self.agent_ext.initialize(self.connection,
                                  lb_agt_constants.EXTENSION_DRIVER_TYPE)
        self.connection.create_consumer.assert_has_calls(
            [mock.call(
                resources_rpc.resource_type_versioned_topic(resource_type),
                [rpc_mock()],
                fanout=True)
             for resource_type in (
                 sfc_obj.BaGPipeChainHop.obj_name(),
                 sfc_obj.BaGPipePortHops.obj_name())],
            any_order=True
        )
        subscribe_mock.assert_has_calls(
            [
                mock.call(mock.ANY, sfc_obj.BaGPipeChainHop.obj_name()),
                mock.call(mock.ANY, sfc_obj.BaGPipePortHops.obj_name())
            ],
            any_order=True
        )

    def _check_port_sfc_info(self, port_id, sides=None):
        self.assertIn(port_id, self.agent_ext.ports_info)
        port_info = self.agent_ext.ports_info[port_id]

        if sides:
            self.assertTrue(port_info.chain_hops)
            for side in sides:
                self.assertTrue(port_info.chain_hops[side])

                hop_keys = list(sfc_obj.BaGPipeChainHop.fields)
                if side == sfc_const.EGRESS:
                    hop_keys += ['lb_consistent_hash_order']
                self.assertTrue(all(key in hop_keys for
                                    key in list(port_info.chain_hops[side])))
        else:
            self.assertFalse(port_info.chain_hops)

    def _fake_chain_hop(self, portchain_id, rts,
                        ingress_network, egress_network,
                        reverse_hop=False,
                        **chain_hop_params):
        chain_hop = dict(
            id=uuidutils.generate_uuid(),
            portchain_id=portchain_id,
            rts=rts,
            ingress_gw=ingress_network['gateway_ip'],
            egress_gw=egress_network['gateway_ip'],
            ingress_ports=net_ports[ingress_network['id']],
            egress_ports=net_ports[egress_network['id']],
            reverse_hop=reverse_hop,
            **chain_hop_params
        )

        return sfc_obj.BaGPipeChainHop(**chain_hop)

    def _chain_hops_notif(self, chain_hops, event_type):
        self.agent_ext.handle_sfc_chain_hops(
            None, sfc_obj.BaGPipeChainHop.obj_name(),
            chain_hops, event_type)

    def _fake_port_hops(self, port_id, ingress_hops=None, egress_hops=None):
        port_hops = dict(
            port_id=port_id,
            ingress_hops=ingress_hops if ingress_hops else [],
            egress_hops=egress_hops if egress_hops else []
        )

        return sfc_obj.BaGPipePortHops(**port_hops)

    def _port_hops_notif(self, port_hops, event_type):
        self.agent_ext.handle_sfc_port_hops(
            None, sfc_obj.BaGPipePortHops.obj_name(),
            port_hops, event_type)

    def test_chain_hop_before_port_up_ingress_only(self):
        ingress_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        self.mocked_bulk_rpc.return_value = [ingress_hop]

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=[]
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'], [sfc_const.INGRESS])

    def test_chain_hop_before_port_up_egress_only(self):
        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        egress_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK2,
            egress_network=base.NETWORK1,
            **egress_params)

        self.mocked_bulk_rpc.return_value = [egress_hop]

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=[],
                        export_rt=CHAIN_HOP_RT1000,
                        readvertise=dict(
                            from_rt=egress_params['readv_from_rts'],
                            to_rt=[egress_params['readv_to_rt']]
                        ),
                        attract_traffic=dict(
                            redirect_rts=egress_params['redirect_rts'],
                            classifier=CHAIN_HOP_CLASSIFIER
                        ),
                        lb_consistent_hash_order=0
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'], [sfc_const.EGRESS])

    def test_two_chain_hops_before_port_up(self):
        ingress_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS2)
        egress_params.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_REVERSE_CLASSIFIER]))
        )

        egress_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT2000,
            ingress_network=base.NETWORK2,
            egress_network=base.NETWORK1,
            **egress_params)

        self.mocked_bulk_rpc.return_value = [ingress_hop, egress_hop]

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=CHAIN_HOP_RT2000,
                        readvertise=dict(
                            from_rt=egress_params['readv_from_rts'],
                            to_rt=[egress_params['readv_to_rt']]
                        ),
                        attract_traffic=dict(
                            redirect_rts=egress_params['redirect_rts'],
                            classifier=CHAIN_HOP_REVERSE_CLASSIFIER
                        ),
                        lb_consistent_hash_order=0
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'],
                                  [sfc_const.INGRESS, sfc_const.EGRESS])

    def test_chain_hop_before_port_delete(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.agent_ext.delete_port(None, self._port_data(base.PORT10,
                                                         delete=True))

        local_port = self._get_expected_local_port(bbgp_const.IPVPN,
                                                   base.NETWORK1['id'],
                                                   base.PORT10['id'],
                                                   detach=True)
        detach_info = {
            'network_id': base.NETWORK1['id'],
            bbgp_const.IPVPN: {
                'ip_address': base.PORT10['ip_address'],
                'mac_address': base.PORT10['mac_address'],
                'local_port': local_port['local_port']
            }
        }

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info)]
        )

        # Verify attachments list consistency
        self.assertEqual(0, len(self.agent_ext.ports_info),
                         "Registered ports list must be empty: %s" %
                         self.agent_ext.ports_info)
        self.assertEqual(0, len(self.agent_ext.networks_info),
                         "Registered networks list must be empty: %s" %
                         self.agent_ext.networks_info)

    def test_chain_hop_created_no_plugged_ports(self):
        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)
        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.assertEqual(0, self.mocked_bagpipe_agent.do_port_plug.call_count,
                         "Do port plug mustn't be called")

    def test_chain_hop_created_already_ingress_plugged_ports(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self.agent_ext.handle_port(None, self._port_data(base.PORT11))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            for port in [base.PORT10, base.PORT11]:
                if port['id'] != port_id:
                    continue

                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK1['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK1['gateway_ip'],
                            local_port=dict(linuxif=linuxbr1),
                            import_rt=CHAIN_HOP_RT1000,
                            export_rt=[]
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port_id in net_ports[base.NETWORK1['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.INGRESS])

    def test_chain_hop_created_already_egress_plugged_ports(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT20))
        self.agent_ext.handle_port(None, self._port_data(base.PORT21))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params.update(
            dict(readv_to_rt=CHAIN_HOP_EGRESS_TO_RT1002,
                 classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr2 = LinuxBridgeManager.get_bridge_name(base.NETWORK2['id'])
            for index, port in enumerate([base.PORT20, base.PORT21]):
                if port['id'] != port_id:
                    continue

                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK2['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK2['gateway_ip'],
                            local_port=dict(linuxif=linuxbr2),
                            import_rt=[],
                            export_rt=CHAIN_HOP_RT1000,
                            readvertise=dict(
                                from_rt=egress_params['readv_from_rts'],
                                to_rt=[egress_params['readv_to_rt']]
                            ),
                            attract_traffic=dict(
                                redirect_rts=egress_params['redirect_rts'],
                                classifier=CHAIN_HOP_CLASSIFIER
                            ),
                            lb_consistent_hash_order=index
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT20['id']), mock.call(base.PORT21['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK2['id'], 2)
        for port_id in net_ports[base.NETWORK2['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.EGRESS])

    def test_chain_hop_created_egress_ports_static_prefix_advertise(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT20))
        self.agent_ext.handle_port(None, self._port_data(base.PORT21))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params.pop('readv_to_rt')
        egress_params.update(
            dict(attract_to_rt=CHAIN_HOP_EGRESS_TO_RT1002,
                 classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr2 = LinuxBridgeManager.get_bridge_name(base.NETWORK2['id'])
            for index, port in enumerate([base.PORT20, base.PORT21]):
                if port['id'] != port_id:
                    continue

                static_prefix = CHAIN_HOP_CLASSIFIER['destinationPrefix']
                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK2['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK2['gateway_ip'],
                            local_port=dict(linuxif=linuxbr2),
                            import_rt=[],
                            export_rt=CHAIN_HOP_RT1000,
                            attract_traffic=dict(
                                to_rt=[egress_params['attract_to_rt']],
                                redirect_rts=egress_params['redirect_rts'],
                                static_destination_prefixes=[static_prefix],
                                classifier=CHAIN_HOP_CLASSIFIER
                            ),
                            lb_consistent_hash_order=index
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT20['id']), mock.call(base.PORT21['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK2['id'], 2)
        for port_id in net_ports[base.NETWORK2['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.EGRESS])

    def test_chain_hop_created_egress_ports_default_prefix_advertise(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT20))
        self.agent_ext.handle_port(None, self._port_data(base.PORT21))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        no_dest_prefix_classifier = copy.copy(CHAIN_HOP_CLASSIFIER)
        no_dest_prefix_classifier.pop('destinationPrefix')

        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params.pop('readv_to_rt')
        egress_params.update(
            dict(attract_to_rt=CHAIN_HOP_EGRESS_TO_RT1002,
                 classifiers=jsonutils.dumps([no_dest_prefix_classifier]))
        )

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr2 = LinuxBridgeManager.get_bridge_name(base.NETWORK2['id'])
            for index, port in enumerate([base.PORT20, base.PORT21]):
                if port['id'] != port_id:
                    continue

                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK2['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK2['gateway_ip'],
                            local_port=dict(linuxif=linuxbr2),
                            import_rt=[],
                            export_rt=CHAIN_HOP_RT1000,
                            attract_traffic=dict(
                                to_rt=[egress_params['attract_to_rt']],
                                redirect_rts=egress_params['redirect_rts'],
                                static_destination_prefixes=['0.0.0.0/0'],
                                classifier=no_dest_prefix_classifier
                            ),
                            lb_consistent_hash_order=index
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT20['id']), mock.call(base.PORT21['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK2['id'], 2)
        for port_id in net_ports[base.NETWORK2['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.EGRESS])

    def test_chain_hop_created_already_plugged_ports(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self.agent_ext.handle_port(None, self._port_data(base.PORT11))

        self.agent_ext.handle_port(None, self._port_data(base.PORT20))
        self.agent_ext.handle_port(None, self._port_data(base.PORT21))

        self.assertEqual(self.mocked_bagpipe_agent.do_port_plug.call_count, 0)

        egress_params = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            for index, port in enumerate([base.PORT10, base.PORT11]):
                if port['id'] != port_id:
                    continue

                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK1['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK1['gateway_ip'],
                            local_port=dict(linuxif=linuxbr1),
                            import_rt=CHAIN_HOP_RT1000,
                            export_rt=[]
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

            linuxbr2 = LinuxBridgeManager.get_bridge_name(base.NETWORK2['id'])
            for index, port in enumerate([base.PORT20, base.PORT21]):
                if port['id'] != port_id:
                    continue

                self.assertDictEqual(
                    dict(
                        network_id=base.NETWORK2['id'],
                        ipvpn=[dict(
                            ip_address=port['ip_address'],
                            mac_address=port['mac_address'],
                            gateway_ip=base.NETWORK2['gateway_ip'],
                            local_port=dict(linuxif=linuxbr2),
                            import_rt=[],
                            export_rt=CHAIN_HOP_RT1000,
                            readvertise=dict(
                                from_rt=egress_params['readv_from_rts'],
                                to_rt=[egress_params['readv_to_rt']]
                            ),
                            attract_traffic=dict(
                                redirect_rts=egress_params['redirect_rts'],
                                classifier=CHAIN_HOP_CLASSIFIER
                            ),
                            lb_consistent_hash_order=index
                        )]
                    ),
                    self.agent_ext.build_sfc_attach_info(port['id'])
                )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id']),
             mock.call(base.PORT20['id']), mock.call(base.PORT21['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port in [base.PORT10, base.PORT11]:
            self._check_port_sfc_info(port['id'], [sfc_const.INGRESS])

        self._check_network_info(base.NETWORK2['id'], 2)
        for port in [base.PORT20, base.PORT21]:
            self._check_port_sfc_info(port['id'], [sfc_const.EGRESS])

    def test_two_chain_hops_created_same_port(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        egress_params_1 = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params_1.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop_1 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params_1)

        egress_params_2 = copy.copy(CHAIN_HOP_EGRESS_PARAMS2)
        egress_params_2.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_REVERSE_CLASSIFIER]))
        )

        chain_hop_2 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT2000,
            ingress_network=base.NETWORK2,
            egress_network=base.NETWORK1,
            **egress_params_2)

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=CHAIN_HOP_RT2000,
                        readvertise=dict(
                            from_rt=egress_params_2['readv_from_rts'],
                            to_rt=[egress_params_2['readv_to_rt']]
                        ),
                        attract_traffic=dict(
                            redirect_rts=egress_params_2['redirect_rts'],
                            classifier=CHAIN_HOP_REVERSE_CLASSIFIER
                        ),
                        lb_consistent_hash_order=0
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._chain_hops_notif([chain_hop_1, chain_hop_2], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'],
                                  [sfc_const.INGRESS, sfc_const.EGRESS])

    def test_one_by_one_chain_hops_created_same_port(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        egress_params_1 = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params_1.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop_1 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params_1)

        # Verify build callback attachments
        def check_build_cb_1(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=[]
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb_1

        self._chain_hops_notif([chain_hop_1], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'], [sfc_const.INGRESS])

        self.mocked_bagpipe_agent.reset_mock()

        egress_params_2 = copy.copy(CHAIN_HOP_EGRESS_PARAMS2)
        egress_params_2.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_REVERSE_CLASSIFIER]))
        )

        chain_hop_2 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT2000,
            ingress_network=base.NETWORK2,
            egress_network=base.NETWORK1,
            **egress_params_2)

        # Verify build callback attachments
        def check_build_cb_2(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=CHAIN_HOP_RT2000,
                        readvertise=dict(
                            from_rt=egress_params_2['readv_from_rts'],
                            to_rt=[egress_params_2['readv_to_rt']]
                        ),
                        attract_traffic=dict(
                            redirect_rts=egress_params_2['redirect_rts'],
                            classifier=CHAIN_HOP_REVERSE_CLASSIFIER
                        ),
                        lb_consistent_hash_order=0
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb_2

        self._chain_hops_notif([chain_hop_2], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'],
                                  [sfc_const.INGRESS, sfc_const.EGRESS])

    def test_port_hop_created_already_plugged_port(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        port10_hops = self._fake_port_hops(base.PORT10['id'],
                                           ingress_hops=[chain_hop])

        # Verify build callback attachments
        def check_build_cb(port_id):
            linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])
            self.assertDictEqual(
                dict(
                    network_id=base.NETWORK1['id'],
                    ipvpn=[dict(
                        ip_address=base.PORT10['ip_address'],
                        mac_address=base.PORT10['mac_address'],
                        gateway_ip=base.NETWORK1['gateway_ip'],
                        local_port=dict(linuxif=linuxbr1),
                        import_rt=CHAIN_HOP_RT1000,
                        export_rt=[]
                    )]
                ),
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
            )

        # we need to check what build_sfc_attach_info returns, at the
        # precise time when do_port_plug is called
        self.mocked_bagpipe_agent.do_port_plug.side_effect = check_build_cb

        self._port_hops_notif([port10_hops], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id'])]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 1)
        self._check_port_sfc_info(base.PORT10['id'], [sfc_const.INGRESS])

    def test_chain_hop_deleted_no_plugged_ports(self):
        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)
        self._chain_hops_notif([chain_hop], rpc_events.DELETED)

        self.mocked_bagpipe_agent.do_port_plug.assert_not_called()
        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_not_called()

    def test_chain_hop_deleted_remaining_ingress_ports(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self.agent_ext.handle_port(None, self._port_data(base.PORT11))

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachment list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port_id in net_ports[base.NETWORK1['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.INGRESS])

        self.mocked_bagpipe_agent.reset_mock()

        # Prepare expected information for DELETE
        linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])

        detach_info10 = dict(
            network_id=base.NETWORK1['id'],
            ipvpn=dict(
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                local_port=dict(linuxif=linuxbr1)
            )
        )

        detach_info11 = dict(
            network_id=base.NETWORK1['id'],
            ipvpn=dict(
                ip_address=base.PORT11['ip_address'],
                mac_address=base.PORT11['mac_address'],
                local_port=dict(linuxif=linuxbr1)
            )
        )

        def check_build_cb(port_id, detach_info):
            self.assertDictEqual(
                {},
                self.agent_ext.build_sfc_attach_info(port_id)
                )

        # we need to check that build_sfc_attach_info contains the expected
        # content precisely at the time when do_port_plug_refresh is called
        self.mocked_bagpipe_agent.do_port_plug_refresh.side_effect = (
            check_build_cb
        )

        # Delete chain hop
        self._chain_hops_notif([chain_hop], rpc_events.DELETED)

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11)],
            any_order=True
        )

        # Verify attachment list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port_id in net_ports[base.NETWORK1['id']]:
            self._check_port_sfc_info(port_id)

    def test_one_by_one_chain_hops_deleted_same_port(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self.agent_ext.handle_port(None, self._port_data(base.PORT11))
        self.agent_ext.handle_port(None, self._port_data(base.PORT20))
        self.agent_ext.handle_port(None, self._port_data(base.PORT21))

        egress_params_1 = copy.copy(CHAIN_HOP_EGRESS_PARAMS1)
        egress_params_1.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_CLASSIFIER]))
        )

        chain_hop_1 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2,
            **egress_params_1)

        egress_params_2 = copy.copy(CHAIN_HOP_EGRESS_PARAMS2)
        egress_params_2.update(
            dict(classifiers=jsonutils.dumps([CHAIN_HOP_REVERSE_CLASSIFIER]))
        )

        chain_hop_2 = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT2000,
            ingress_network=base.NETWORK2,
            egress_network=base.NETWORK1,
            **egress_params_2)

        self._chain_hops_notif([chain_hop_1, chain_hop_2], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id']),
             mock.call(base.PORT20['id']), mock.call(base.PORT21['id'])],
            any_order=True
        )

        # Verify attachment list consistency
        for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
            self._check_network_info(network_id, 2)
            for port_id in net_ports[network_id]:
                self._check_port_sfc_info(port_id,
                                          [sfc_const.INGRESS,
                                           sfc_const.EGRESS])

        self.mocked_bagpipe_agent.reset_mock()

        # Prepare expected information for DELETE
        linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])

        detach_info10 = dict(
            network_id=base.NETWORK1['id'],
            ipvpn=dict(
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                local_port=dict(linuxif=linuxbr1)
            )
        )

        detach_info11 = dict(
            network_id=base.NETWORK1['id'],
            ipvpn=dict(
                ip_address=base.PORT11['ip_address'],
                mac_address=base.PORT11['mac_address'],
                local_port=dict(linuxif=linuxbr1)
            )
        )

        linuxbr2 = LinuxBridgeManager.get_bridge_name(base.NETWORK2['id'])

        detach_info20 = dict(
            network_id=base.NETWORK2['id'],
            ipvpn=dict(
                ip_address=base.PORT20['ip_address'],
                mac_address=base.PORT20['mac_address'],
                local_port=dict(linuxif=linuxbr2)
            )
        )

        detach_info21 = dict(
            network_id=base.NETWORK2['id'],
            ipvpn=dict(
                ip_address=base.PORT21['ip_address'],
                mac_address=base.PORT21['mac_address'],
                local_port=dict(linuxif=linuxbr2)
            )
        )

        self.port2build_attach = {
            base.PORT10['id']: dict(
                network_id=base.NETWORK1['id'],
                ipvpn=[dict(
                    ip_address=base.PORT10['ip_address'],
                    mac_address=base.PORT10['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=linuxbr1),
                    import_rt=CHAIN_HOP_RT1000,
                    export_rt=[]
                )]
            ),
            base.PORT11['id']: dict(
                network_id=base.NETWORK1['id'],
                ipvpn=[dict(
                    ip_address=base.PORT11['ip_address'],
                    mac_address=base.PORT11['mac_address'],
                    gateway_ip=base.NETWORK1['gateway_ip'],
                    local_port=dict(linuxif=linuxbr1),
                    import_rt=CHAIN_HOP_RT1000,
                    export_rt=[]
                )]
            ),
            base.PORT20['id']: dict(
                network_id=base.NETWORK2['id'],
                ipvpn=[dict(
                    ip_address=base.PORT20['ip_address'],
                    mac_address=base.PORT20['mac_address'],
                    gateway_ip=base.NETWORK2['gateway_ip'],
                    local_port=dict(linuxif=linuxbr2),
                    import_rt=[],
                    export_rt=CHAIN_HOP_RT1000,
                    readvertise=dict(
                        from_rt=egress_params_1['readv_from_rts'],
                        to_rt=[egress_params_1['readv_to_rt']]
                    ),
                    attract_traffic=dict(
                        redirect_rts=egress_params_1['redirect_rts'],
                        classifier=CHAIN_HOP_CLASSIFIER
                    ),
                    lb_consistent_hash_order=0
                )]
            ),
            base.PORT21['id']: dict(
                network_id=base.NETWORK2['id'],
                ipvpn=[dict(
                    ip_address=base.PORT21['ip_address'],
                    mac_address=base.PORT21['mac_address'],
                    gateway_ip=base.NETWORK2['gateway_ip'],
                    local_port=dict(linuxif=linuxbr2),
                    import_rt=[],
                    export_rt=CHAIN_HOP_RT1000,
                    readvertise=dict(
                        from_rt=egress_params_1['readv_from_rts'],
                        to_rt=[egress_params_1['readv_to_rt']]
                    ),
                    attract_traffic=dict(
                        redirect_rts=egress_params_1['redirect_rts'],
                        classifier=CHAIN_HOP_CLASSIFIER
                    ),
                    lb_consistent_hash_order=1
                )]
            )
        }

        def check_build_cb_1(port_id, detach_info):
            self.assertDictEqual(
                self.port2build_attach[port_id],
                self.agent_ext.build_sfc_attach_info(port_id)
            )

        # we need to check that build_sfc_attach_info contains the expected
        # content precisely at the time when do_port_plug_refresh is called
        self.mocked_bagpipe_agent.do_port_plug_refresh.side_effect = (
            check_build_cb_1
        )

        self._chain_hops_notif([chain_hop_2], rpc_events.DELETED)

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11),
             mock.call(base.PORT20['id'], detach_info20),
             mock.call(base.PORT21['id'], detach_info21)],
            any_order=True
        )

        # Verify attachment list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port in [base.PORT10, base.PORT11]:
            self._check_port_sfc_info(port['id'], [sfc_const.INGRESS])

        self._check_network_info(base.NETWORK2['id'], 2)
        for port in [base.PORT20, base.PORT21]:
            self._check_port_sfc_info(port['id'], [sfc_const.EGRESS])

        self.mocked_bagpipe_agent.reset_mock()

        def check_build_cb_2(port_id, detach_info):
            self.assertDictEqual(
                {},
                self.agent_ext.build_sfc_attach_info(port_id)
                )

        # we need to check that build_sfc_attach_info contains the expected
        # content precisely at the time when do_port_plug_refresh is called
        self.mocked_bagpipe_agent.do_port_plug_refresh.side_effect = (
            check_build_cb_2
        )

        # Delete port hop
        self._chain_hops_notif([chain_hop_1], rpc_events.DELETED)

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10),
             mock.call(base.PORT11['id'], detach_info11),
             mock.call(base.PORT20['id'], detach_info20),
             mock.call(base.PORT21['id'], detach_info21)],
            any_order=True
        )

        # Verify attachment list consistency
        for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
            self._check_network_info(network_id, 2)
            for port_id in net_ports[network_id]:
                self._check_port_sfc_info(port_id)

    def test_port_hop_deleted_remaining_ingress_ports(self):
        self.agent_ext.handle_port(None, self._port_data(base.PORT10))
        self.agent_ext.handle_port(None, self._port_data(base.PORT11))

        chain_hop = self._fake_chain_hop(
            portchain_id=uuidutils.generate_uuid(),
            rts=CHAIN_HOP_RT1000,
            ingress_network=base.NETWORK1,
            egress_network=base.NETWORK2)

        self._chain_hops_notif([chain_hop], rpc_events.CREATED)

        self.mocked_bagpipe_agent.do_port_plug.assert_has_calls(
            [mock.call(base.PORT10['id']), mock.call(base.PORT11['id'])],
            any_order=True
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        for port_id in net_ports[base.NETWORK1['id']]:
            self._check_port_sfc_info(port_id, [sfc_const.INGRESS])

        self.mocked_bagpipe_agent.reset_mock()

        # Prepare expected information for DELETE
        linuxbr1 = LinuxBridgeManager.get_bridge_name(base.NETWORK1['id'])

        detach_info10 = dict(
            network_id=base.NETWORK1['id'],
            ipvpn=dict(
                ip_address=base.PORT10['ip_address'],
                mac_address=base.PORT10['mac_address'],
                local_port=dict(linuxif=linuxbr1)
            )
        )

        def check_build_cb(port_id, detach_info):
            self.assertDictEqual(
                {},
                self.agent_ext.build_sfc_attach_info(base.PORT10['id'])
                )

        # we need to check that build_sfc_attach_info contains the expected
        # content precisely at the time when do_port_plug_refresh is called
        self.mocked_bagpipe_agent.do_port_plug_refresh.side_effect = (
            check_build_cb
        )

        # Delete port hop
        port10_hops = self._fake_port_hops(base.PORT10['id'],
                                           ingress_hops=[chain_hop])
        self._port_hops_notif([port10_hops], rpc_events.DELETED)

        self.mocked_bagpipe_agent.do_port_plug_refresh.assert_has_calls(
            [mock.call(base.PORT10['id'], detach_info10)]
        )

        # Verify attachments list consistency
        self._check_network_info(base.NETWORK1['id'], 2)
        self._check_port_sfc_info(base.PORT10['id'])
        self._check_port_sfc_info(base.PORT11['id'], [sfc_const.INGRESS])



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from networking_bagpipe.bagpipe_bgp.common import dataplane_utils
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.bagpipe_bgp.vpn.evpn import ovs
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t

from neutron.agent.common import ovs_lib
from neutron.plugins.ml2.drivers.openvswitch.agent.common import \
    constants as ovs_const
from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.ovs_ofctl import \
    br_tun

LOCAL_IP = "1.2.3.4"
MAC1 = "01:00:de:ad:be:ef"
MAC2 = "01:00:fe:ed:f0:0d"


class TestTunnelManager(t.TestCase):

    def setUp(self):
        super(TestTunnelManager, self).setUp()

        self.bridge = mock.Mock(spec=br_tun.OVSTunnelBridge)
        self.manager = ovs.TunnelManager(self.bridge, LOCAL_IP)

    def test_get_tunnel(self):
        t1 = self.manager.get_object("2.2.2.2", "A")
        self.bridge.add_tunnel_port.assert_called_once_with(mock.ANY,
                                                            "2.2.2.2",
                                                            LOCAL_IP,
                                                            mock.ANY)
        self.bridge.setup_tunnel_port.assert_called_once_with(mock.ANY, t1)

        self.bridge.add_tunnel_port.reset_mock()
        self.bridge.setup_tunnel_port.reset_mock()

        t2 = self.manager.get_object("2.2.2.2", "B")
        self.bridge.add_tunnel_port.assert_not_called()
        self.bridge.setup_tunnel_port.assert_not_called()

        self.assertTrue(t1 == t2)

        self.bridge.add_tunnel_port.reset_mock()
        self.bridge.setup_tunnel_port.reset_mock()

        t3 = self.manager.get_object("3.3.3.3", "A")
        self.bridge.add_tunnel_port.assert_called_once()
        self.bridge.setup_tunnel_port.assert_called_once()

        self.assertFalse(t3 != t1)

        self.assertTrue(len(self.manager.infos()))

    def test_free_object(self):
        t1 = self.manager.get_object("2.2.2.2", "A")
        self.manager.get_object("2.2.2.2", "B")
        t2 = self.manager.get_object("3.3.3.3", "A")

        self.bridge.add_tunnel_port.reset_mock()
        self.bridge.delete_port.reset_mock()
        self.bridge.setup_tunnel_port.reset_mock()

        self.manager.free_object("2.2.2.2", "A")
        self.bridge.delete_port.assert_not_called()
        t1bis = self.manager.get_object("2.2.2.2")
        self.assertTrue(t1bis == t1)

        self.bridge.add_tunnel_port.reset_mock()
        self.bridge.delete_port.reset_mock()
        self.bridge.setup_tunnel_port.reset_mock()

        self.manager.free_object("2.2.2.2", "B")
        self.bridge.delete_port.assert_called_once_with(t1)

        self.bridge.add_tunnel_port.reset_mock()
        self.bridge.delete_port.reset_mock()
        self.bridge.setup_tunnel_port.reset_mock()

        self.manager.free_object("3.3.3.3", "A")
        self.bridge.delete_port.assert_called_once_with(t2)


class FakeBridgeMockSpec(dataplane_utils.OVSBridgeWithGroups,
                         br_tun.OVSTunnelBridge,
                         ovs_lib.OVSBridge):
    pass


class FakeNLRI(object):

    def __init__(self, ip):
        self.ip = ip


class TestOVSEVIDataplane(t.TestCase):

    def setUp(self):
        super(TestOVSEVIDataplane, self).setUp()

        self.bridge = mock.Mock(spec=FakeBridgeMockSpec)
        self.tunnel_mgr = mock.Mock(spec=ovs.TunnelManager)

        self.dp_driver = mock.Mock(spec=dp_drivers.DataplaneDriver)
        self.dp_driver.bridge = self.bridge
        self.dp_driver.tunnel_mgr = self.tunnel_mgr
        self.dp_driver.config = {}
        self.dp_driver.get_local_address.return_value = LOCAL_IP

        self.dataplane = ovs.OVSEVIDataplane(
            self.dp_driver, 77, "foo_external_instance_id",
            None, None, instance_label=99)
        self.vlan = 99

        self.dataplane.vif_plugged("MAC1", "10.0.0.1",
                                   {'vlan': self.vlan},
                                   None, None)
        self.bridge.add_flow.assert_called_once_with(
            table=ovs_const.VXLAN_TUN_TO_LV,
            priority=mock.ANY,
            tun_id=99,
            actions=mock.ANY)

    def test_setup_dataplane_for_remote_endpoint__local(self):
        self.dataplane.setup_dataplane_for_remote_endpoint(
            MAC1, LOCAL_IP, 42, FakeNLRI("11.0.0.1"), None)

        self.tunnel_mgr.get_object.assert_not_called()

    def test_setup_dataplane_for_remote_endpoint(self):
        self.dataplane.setup_dataplane_for_remote_endpoint(
            MAC1, "2.2.2.2", 42, FakeNLRI("11.0.0.1"), None)

        self.tunnel_mgr.get_object.assert_called_once_with(
            "2.2.2.2", (77, (42, MAC1)))

        self.bridge.add_flow.assert_called_with(
            table=ovs_const.UCAST_TO_TUN,
            priority=mock.ANY,
            dl_vlan=self.vlan,
            dl_dst=MAC1,
            actions=mock.ANY)

    def test_remove_dataplane_for_remote_endpoint__local(self):
        self.dataplane.remove_dataplane_for_remote_endpoint(
            MAC1, LOCAL_IP, 42, FakeNLRI("11.0.0.1"))

        self.bridge.delete_unicast_to_tun.assert_called_with(self.vlan, MAC1)

        self.tunnel_mgr.free_object.assert_not_called()

    def test_remove_dataplane_for_remote_endpoint(self):
        self.dataplane.remove_dataplane_for_remote_endpoint(
            MAC1, "2.2.2.2", 42, FakeNLRI("11.0.0.1"))

        self.bridge.delete_unicast_to_tun.assert_called_with(self.vlan, MAC1)

        self.tunnel_mgr.free_object.assert_called_with(
            "2.2.2.2", (77, (42, MAC1)))

    def test_add_dataplane_for_bum_endpoint__local(self):
        self.dataplane.add_dataplane_for_bum_endpoint(LOCAL_IP, 45,
                                                      None, None)

        self.tunnel_mgr.get_object.assert_not_called()

    def test_add_dataplane_for_bum_endpoint(self):
        self.dataplane.add_dataplane_for_bum_endpoint("2.2.2.2", 45,
                                                      None, None)

        self.tunnel_mgr.get_object.assert_called_with(
            "2.2.2.2", (77, (45, "flood")))

    def test_remove_dataplane_for_bum_endpoint__local(self):
        self.dataplane.add_dataplane_for_bum_endpoint(LOCAL_IP, 45,
                                                      None, None)
        self.tunnel_mgr.free_object.reset_mock()

        self.dataplane.remove_dataplane_for_bum_endpoint(LOCAL_IP, 45, None)

        self.tunnel_mgr.free_object.assert_not_called()

    def test_remove_dataplane_for_bum_endpoint(self):
        self.dataplane.add_dataplane_for_bum_endpoint("2.2.2.2", 45,
                                                      None, None)
        self.tunnel_mgr.free_object.reset_mock()

        self.dataplane.remove_dataplane_for_bum_endpoint("2.2.2.2", 45, None)

        self.tunnel_mgr.free_object.assert_called_with(
            "2.2.2.2", (77, (45, "flood")))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2017 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging as python_logging
import time
import uuid

from oslo_config import cfg
from oslo_log import log as logging
import pbr.version
import pecan
from pecan import request

from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.vpn import manager as vpn_manager


LOG = logging.getLogger(__name__)

LOOKING_GLASS_BASE = "looking-glass"


def expose(*args, **kwargs):
    """Helper function so we don't have to specify json for everything."""
    kwargs.setdefault('content_type', 'application/json')
    kwargs.setdefault('template', 'json')
    return pecan.expose(*args, **kwargs)


def when(index, *args, **kwargs):
    """Helper function so we don't have to specify json for everything."""
    kwargs.setdefault('content_type', 'application/json')
    kwargs.setdefault('template', 'json')
    return index.when(*args, **kwargs)


class PingController(object):

    def __init__(self):
        # Random generated sequence number
        self.sequence = int(uuid.uuid4())

    @expose(generic=True)
    def index(self):
        return self.sequence


class VPNManagerController(object):

    def __init__(self):
        self.manager = vpn_manager.VPNManager.get_instance()

    @staticmethod
    def stop():
        vpn_manager.VPNManager.get_instance().stop()


class AttachController(VPNManagerController):
    """attach_localport parameters:

    'vpn_instance_id': external VPN instance identifier (all ports with same
                     vpn_instance_id will be plugged in the same VPN
                     instance
    'instance_description': free form description of VPN instance
    'description': free form description of attachment
    'vpn_type': type of the VPN instance ('ipvpn' or 'evpn')
    'import_rt': list of import Route Targets (or comma-separated string)
    'export_rt': list of export Route Targets (or comma-separated string)
    'gateway_ip': IP address of gateway for this VPN instance
    'mac_address': MAC address of endpoint to connect to the VPN instance
    'ip_address': IP/mask of endpoint to connect to the VPN instance
    'advertise_subnet': optional, if set to True then VRF will advertise
                        the whole subnet (defaults to False, readvertise
                        ip_address as a singleton (/32)
    'linuxbr': Name of a linux bridge to which the linuxif is already
             plugged-in (optional)
    'vni': VXLAN VNI to use (optional)
    'local_pref': BGP LOCAL_PREF for the route to this vif (optional)
    'direction': 'to-port' | 'from-port' | 'both'
        # specify local port traffic direction in VPN instance
        # (route advertisements are not done with from-port only)
        # to-port: only forward traffic to the VIF
        # from-port: only forward traffic to the VIF
    'local_port': local port to plug to the VPN instance
        should be a dict containing any of the following key,value pairs
        {
            'linuxif': 'tap456abc', # name of a linux interface
                                    # - if OVS information is provided it
                                    #   does not have to be an existing
                                    #   interface
                                    # - not needed/not used if 'evpn' plug
                                    #   is used
            'ovs': {  # optional
                # whether or not interface is already plugged into the
                # OVS bridge:
                'plugged': True,
                # name of a linux interface to be plugged into the OVS
                # bridge (optional and ignored if port_number is
                # provided):
                'port_name': 'qvo456abc',
                # OVS port number (optional if 'port_name' provided):
                'port_number': '7',
                # the VLAN id for VM traffic (optional)
                'vlan': '42',
                # optional specification of a distinct port to send
                # traffic to the VM(only applies if a vlan is
                # specified) :
                'to_vm_port_number'
                'to_vm_port_name'
               },
            'evpn': {  # for an ipvpn attachment...
                 'id': 'xyz'  # specifies the vpn_instance_id of an evpn
                              # that will be attached to the ipvpn
                 'ovs_port_name': 'qvb456abc' # optional, if provided,
                                              # and if ovs/port_name is
                                              # also provided, then the
                                              # interface name will be
                                              # assumed as already plugged
                                              # into the evpn
                }
        }
        if local_port is not a list, it is assumed to be a name of a linux
        interface (string)
    'readvertise': {  # optional, used to re-advertise addresses...
        'from_rt': [list of RTs]  # ...received on these RTs
        'to_rt': [list of RTs] # ...toward these RTs
    }
    'attract_traffic': { # optional, will result in the generation of FlowSpec
                         # routes, based on the specified classifier,
                         # advertised to the readvertise:to_rt RTs, redirecting
                         # traffic toward the "attract_to_rts" RT using a
                         # redirect-to-VRF action action.
                         # The prefixes for which FlowSpec routes are
                         # advertised are:
                         # - prefixes carrying one of the readvertise:from_rt
                         #   RTs,
                         # - prefixes in static_destination_prefixes.
                         #
                         # When this is used, the routes that are advertised to
                         # the readvertise:to_rt route targets are, instead of
                         # the prefixes of the routes carrying an RT in
                         # readvertise:from_rt, the prefix in
                         # static_destination_prefixes if any, or else a
                         # default0.0.0.0 route. These routes are advertised
                         # for each locally attached interface, each time with
                         # a distinct RD
        'classifier': {
            'sourcePrefix': IP/mask,
            'sourcePort': Port number or port range,
            'destinationPort': Port number or port range,
            'protocol': IP protocol
        },
        'redirect_rts': [list of RTs] # RTs of generated FlowSpec routes
        'attract_to_rts': [list of RTs] # RTs of the redirect-to-VRF action of
                                        # the generated FlowSpec routes
        'static_destination_prefixes': [list of IP/mask] # When present,
                                                         # FlowSpec routes will
                                                         # be generated from
                                                         # these prefixes (as
                                                         # destination prefix);
                                                         # this is done
                                                         # additionally to
                                                         # FlowSpec routes
                                                         # generated from
                                                         # routes carrying
                                                         # readvertise:from_rt
    }
    'fallback': # (optional) if provided, on a VRF lookup miss,
                # the MAC destination address will be
                # rewritten to this MAC before being
                # sent back where it came from
                {
                'src_mac': 'aa:bb:cc:dd:ee:ff'  # new source MAC
                'dst_mac': 'aa:bb:cc:dd:ee:00'  # new destination MAC
                'ovs_port_name': 'patch_foo'
                'ovs_port_number': 4               # (unsupported yet)
                'ovs_resubmit': '(<port>,<table>)' # (unsupported yet)
    }
    """

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    @when(index, method='POST')
    def process(self):
        try:
            attach_params = request.json
        except Exception:
            LOG.error('attach_localport: No local port details received')
            pecan.abort(400, 'No local port details received')

        try:
            LOG.info('Local port attach received: %s', attach_params)

            self.manager.plug_vif_to_vpn(**attach_params)
        except exc.APIException as e:
            LOG.warning('attach_localport: API error: %s', e)
            pecan.abort(400, "API error: %s" % e)
        except Exception:
            LOG.exception('attach_localport: An error occurred during local '
                          'port plug to VPN')
            pecan.abort(500, 'An error occurred during local port plug to VPN')


class DetachController(VPNManagerController):

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    @when(index, method='POST')
    def process(self):
        try:
            detach_params = request.json
        except Exception:
            LOG.error('detach_localport: No local port details received')
            pecan.abort(400, 'No local port details received')

        try:
            LOG.info('Local port detach received: %s', detach_params)
            self.manager.unplug_vif_from_vpn(**detach_params)
        except exc.APIException as e:
            LOG.warning('detach_localport: API error: %s', e)
            pecan.abort(400, "API error: %s" % e)
        except Exception:
            LOG.exception('detach_localport: An error occurred during local '
                          'port unplug from VPN')
            pecan.abort(500, 'An error occurred during local port unplug from '
                        'VPN')


class LookingGlassController(VPNManagerController,
                             lg.LookingGlassMixin):

    def __init__(self):
        super(LookingGlassController, self).__init__()

        self.start_time = time.time()

        lg.set_references_root(LOOKING_GLASS_BASE)
        lg.set_reference_path("BGP_WORKERS", ["bgp", "workers"])
        lg.set_reference_path("VPN_INSTANCES", ["vpns", "instances"])
        lg.set_reference_path("DATAPLANE_DRIVERS",
                              ["vpns", "dataplane", "drivers"])

        self.catchall_lg_log_handler = lg.LookingGlassLogHandler()
        python_logging.getLogger().addHandler(self.catchall_lg_log_handler)

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='GET')
    def process(self, *url_path_elements):

        path_prefix = "%s://%s/%s" % (
            request.scheme,  # http
            request.host,
            LOOKING_GLASS_BASE,
        )

        try:
            lg_info = self.get_looking_glass_info(path_prefix,
                                                  url_path_elements)
            if lg_info is None:
                raise lg.NoSuchLookingGlassObject(path_prefix,
                                                  url_path_elements[0])
            return lg_info
        except lg.NoSuchLookingGlassObject as e:
            LOG.info('looking_glass: %s', repr(e))
            pecan.abort(404, repr(e))
        except Exception:
            LOG.exception('looking_glass: an error occurred')
            pecan.abort(500, 'Server error')

    @when(index, method='DELETE')
    @when(index, method='POST')
    @when(index, method='PUT')
    def not_supported(self):
        pecan.abort(405)

    # Looking glass hooks #################

    def get_lg_map(self):
        return {
            "summary":  (lg.SUBITEM, self.get_lg_summary),
            "config":   (lg.SUBTREE, self.get_lg_config),
            "bgp":      (lg.DELEGATE, self.manager.bgp_manager),
            "vpns":     (lg.DELEGATE, self.manager),
            "logs":     (lg.SUBTREE, self.get_logs)
        }

    def get_lg_config(self, path_prefix):
        return {section: utils.osloconfig_json_serialize(cfg.CONF[section])
                for section in ('COMMON', 'API', 'BGP',
                                'DATAPLANE_DRIVER_IPVPN',
                                'DATAPLANE_DRIVER_EVPN')
                }

    def get_lg_summary(self):
        return {
            "BGP_established_peers":
                self.manager.bgp_manager.get_established_peers_count(),
            "local_routes_count":
                self.manager.bgp_manager.rtm.
                get_local_routes_count(),
            "received_routes_count":
                self.manager.bgp_manager.rtm.
                get_received_routes_count(),
            "vpn_instances_count": self.manager.get_vpn_instances_count(),
            "warnings_and_errors": len(self.catchall_lg_log_handler),
            "start_time": time.strftime("%Y-%m-%d %H:%M:%S",
                                        time.localtime(self.start_time)),
            "version":  (pbr.version.VersionInfo('networking-bagpipe')
                         .release_string())
        }

    def get_logs(self, path_prefix):
        return [{'level': record.levelname,
                 'time':
                 self.catchall_lg_log_handler.formatter.formatTime(record),
                 'name': record.name,
                 'message': record.msg}
                for record in self.catchall_lg_log_handler.get_records()]


class RootController(object):

    @expose(generic=True)
    def index(self):
        return {}

    @when(index, method='POST')
    @when(index, method='PUT')
    @when(index, method='DELETE')
    def not_supported(self):
        pecan.abort(405)

    ping = PingController()
    attach_localport = AttachController()
    detach_localport = DetachController()

    def stop(self):
        VPNManagerController.stop()


# there is a '-' in the LOOKING_GLASS_BASE name, so we have to use pecan.route
pecan.route(RootController, LOOKING_GLASS_BASE, LookingGlassController())



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module classes related to producing and consuming events related to
BGP routes.

routes: RouteEntry

events: RouteEvent
    an announcement or a withdrawal of a BGP route

workers: Worker
    * produce events
    * subscribe to the route table manager to consume events related to
      certain BGP routes

route table manager (singleton)
    * tracks subscriptions of workers
    * dispatches events based on subscriptions

"""

from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.engine import exa


LOG = logging.getLogger(__name__)


class RouteEntry(lg.LookingGlassMixin):
    """A route entry describes a BGP route

    i.e. the association of:
    * a BGP NLRI of a specific type (e.g. a VPNv4 route
      like "1.2.3.4:5:192.168.0.5/32")
    * BGP attributes
    * the source of the BGP route (e.g. the BGP peer, or the local VPN
      instance, that advertizes the route)
    """

    def __init__(self, nlri, rts=None, attributes=None,
                 source=None):
        if attributes is None:
            attributes = exa.Attributes()
        assert isinstance(attributes, exa.Attributes)
        if rts is not None:
            assert isinstance(rts, list)
            assert len(rts) == 0 or isinstance(rts[0], exa.RouteTarget)

        self.source = source
        self.afi = nlri.afi
        self.safi = nlri.safi
        self.nlri = nlri
        self.attributes = attributes
        # a list of exa.bgp.message.update.attribute.community.
        #   extended.RouteTargetASN2Number
        self._route_targets = []
        if exa.Attribute.CODE.EXTENDED_COMMUNITY in self.attributes:
            ecoms = self.attributes[
                exa.Attribute.CODE.EXTENDED_COMMUNITY].communities
            # use type(..) because isinstance(rtrecord, RouteTarget) is True
            self._route_targets = [ecom for ecom in ecoms
                                   if type(ecom) == exa.RouteTarget]
            if rts:
                ecoms += rts
                self._route_targets += rts
        else:
            if rts:
                self.attributes.add(exa.ExtendedCommunities(rts))
                self._route_targets += rts

    @property
    def route_targets(self):
        return self._route_targets

    def ecoms(self, filter_=None):
        '''returns the route extended communities (+ optional filtering)'''
        if filter_ is None:
            def filter_real(ecom):
                return True
        elif isinstance(filter_, six.class_types):
            def filter_real(ecom):
                return isinstance(ecom, filter_)
        else:
            # filter is a function(ecom)
            filter_real = filter_

        if exa.Attribute.CODE.EXTENDED_COMMUNITY in self.attributes:
            return list(
                filter(filter_real,
                       self.attributes[exa.Attribute.CODE.EXTENDED_COMMUNITY]
                           .communities)
                )
        else:
            return []

    @log_decorator.log
    def set_route_targets(self, route_targets):
        # first build a list of ecoms without any RT
        ecoms = self.ecoms(lambda ecom: not isinstance(ecom,
                                                       exa.RouteTarget))

        # then add the right RTs
        new_ecoms = exa.ExtendedCommunities()
        new_ecoms.communities += ecoms
        new_ecoms.communities += route_targets

        # update
        self._route_targets = route_targets

        self.attributes.remove(new_ecoms.ID)
        self.attributes.add(new_ecoms)

    @property
    def nexthop(self):
        try:
            return self.nlri.nexthop.top()
        except AttributeError:
            try:
                return self.attributes[exa.Attribute.CODE.NEXT_HOP].top()
            except KeyError:
                return None

    def __eq__(self, other):
        if other is None:
            return False
        assert isinstance(other, RouteEntry)
        return (self.afi == other.afi and
                self.safi == other.safi and
                self.source == other.source and
                self.nlri == other.nlri and
                self.attributes.sameValuesAs(other.attributes))

    def __hash__(self):
        return hash((self.afi, self.safi, str(self.source),
                     str(self.nexthop), self.nlri,
                     self.attributes))

    def __repr__(self, skip_nexthop=False):
        from_string = " from:%s" % self.source if self.source else ""

        nexthop = ""
        if not skip_nexthop:
            nexthop = str(self.nexthop)

        return "[RouteEntry: %s/%s %s nh:%s %s%s]" % (
            self.afi, self.safi, self.nlri, nexthop,
            self.attributes, from_string)

    def get_lg_local_info(self, path_prefix):

        att_dict = {}

        for attribute in self.attributes.values():

            # skip some attributes that we care less about
            if attribute.ID in (exa.Attribute.CODE.AS_PATH,
                                exa.Attribute.CODE.ORIGIN):
                continue

            att_dict[
                repr(exa.Attribute.CODE(attribute.ID))] = str(attribute)

        res = {"afi-safi": "%s/%s" % (self.afi, self.safi),
               "attributes": att_dict,
               "next_hop": self.nexthop
               }

        if self.source:
            res["source"] = {"id": self.source.name,
                             "href": lg.get_absolute_path("BGP_WORKERS",
                                                          path_prefix,
                                                          [self.source.name])
                             }

        if self.safi in [exa.SAFI.mpls_vpn, exa.SAFI.evpn]:
            res["route_targets"] = [str(rt) for rt in self.route_targets]

        return {
            repr(self.nlri): res
        }


class RouteEvent(object):

    """Represents an advertisement or withdrawal of a RouteEntry"""

    # event Types
    ADVERTISE = 1
    WITHDRAW = 2

    type2name = {ADVERTISE: "Advertise",
                 WITHDRAW: "Withdraw"}

    def __init__(self, event_type, route_entry, source=None):
        assert(event_type == RouteEvent.ADVERTISE or
               event_type == RouteEvent.WITHDRAW)
        assert isinstance(route_entry, RouteEntry)
        self.type = event_type
        self.route_entry = route_entry
        if source is not None:
            self.source = source
            self.route_entry.source = source
        else:
            self.source = route_entry.source
        assert self.source is not None
        self.replaced_route = None

        # this is required to overwrite the action field in an NLRI
        # in the case where we generate a withdraw from an existing NLRI
        # on a replaced route
        # and this spares us the pain of specifying the action
        # when creating an nlri
        if event_type == RouteEvent.ADVERTISE:
            self.route_entry.nlri.action = exa.OUT.ANNOUNCE
        else:  # WITHDRAW
            self.route_entry.nlri.action = exa.OUT.WITHDRAW

    @log_decorator.log
    def set_replaced_route(self, replaced_route):
        '''set_replaced_route hook for RouteTableManager

        Called only by RouteTableManager.
        replaced_route should be a RouteEntry
        '''
        assert(isinstance(replaced_route, RouteEntry)
               or (replaced_route is None))
        assert replaced_route != self.route_entry
        self.replaced_route = replaced_route

    def __repr__(self):
        if self.replaced_route:
            replaces_str = "replaces one route"
        else:
            replaces_str = "replaces no route"
        return "[RouteEvent(%s): %s %s %s]" % (replaces_str,
                                               RouteEvent.type2name[self.type],
                                               self.route_entry,
                                               self.source)


class _SubUnsubCommon(object):

    def __init__(self, afi, safi, route_target, worker=None):
        assert route_target is None or isinstance(route_target,
                                                  exa.RouteTarget)
        self.afi = afi
        self.safi = safi
        self.route_target = route_target
        self.worker = worker

    def __repr__(self):
        by_worker = " by %s" % self.worker.name if self.worker else ""
        return "%s [%s/%s,%s]%s" % (self.__class__.__name__,
                                    self.afi or "*", self.safi or "*",
                                    self.route_target or "*", by_worker)


class Subscription(_SubUnsubCommon):

    """Represents a Subscription to RouteEvents

    A subscription specifies the AFI, the SAFI, and the Route Target of the
    RouteEntry for which the subscriber wants to receive events.

    Any of these (afi, safi or route target) can be replaced by a wildcard:
    * Subscription.ANY_AFI
    * Subscription.ANY_SAFI
    * Subscription.ANY_RT
    """

    ANY_AFI = None
    ANY_SAFI = None
    ANY_RT = None

    def __init__(self, afi, safi, route_target=None, worker=None):
        _SubUnsubCommon.__init__(self, afi, safi, route_target, worker)


class Unsubscription(_SubUnsubCommon):

    def __init__(self, afi, safi, route_target=None, worker=None):
        _SubUnsubCommon.__init__(self, afi, safi, route_target, worker)


class EventSource(lg.LookingGlassMixin):
    '''Class for objects that advertise and withdraw routes

    class needs to have a 'name' attribute
    '''

    def __init__(self, route_table_manager):
        self.rtm = route_table_manager
        # private data of RouteTableManager
        self._rtm_route_entries = set()

    def get_route_entries(self):
        return self._rtm_route_entries

    @log_decorator.log_info
    def _advertise_route(self, route_entry):
        LOG.debug("Publish advertise route event")
        self.rtm.enqueue(RouteEvent(RouteEvent.ADVERTISE, route_entry, self))

    @log_decorator.log_info
    def _withdraw_route(self, route_entry):
        LOG.debug("Publish withdraw route event")
        self.rtm.enqueue(RouteEvent(RouteEvent.WITHDRAW, route_entry, self))

    def get_lg_map(self):
        return {
            "adv_routes": (lg.SUBTREE, self.get_lg_routes)
        }

    def get_lg_routes(self, path_prefix):
        return [route.get_looking_glass_info(path_prefix) for route in
                self.get_route_entries()]


class WorkerCleanupEvent(object):

    def __init__(self, worker):
        self.worker = worker

    def __repr__(self):
        return "WorkerCleanupEvent:%s" % (self.worker.name)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import re

from oslo_concurrency import lockutils
from oslo_config import cfg

from networking_bagpipe.bagpipe_bgp.engine import exa


def synchronized(method):

    def synchronized_method(self, *arg, **kws):
        with self.lock:
            return method(self, *arg, **kws)

    return synchronized_method


oslo_synchronized = lockutils.synchronized_with_prefix('bagpipe-bgp-')


def plural(x):
    if len(x) > 1:
        return "s"
    else:
        return ""


def invert_dict_of_sets(d):
    '''return inverted dict of sets from original dict

    original dict possibly containing sets of non-unique hashable items
    '''
    new_d = collections.defaultdict(set)
    for k in d:
        for v in d[k]:
            new_d[v].add(k)
    return new_d


camel2underscore_regex = re.compile('(?!^)([A-Z]+)')


def dict_camelcase_to_underscore(dictionary):
    '''copy dict, with translation of keys from FooBar to foo_bar'''
    return {camel2underscore_regex.sub(r'_\1', key).lower(): value
            for (key, value) in dictionary.items()
            }


def osloconfig_json_serialize(obj):
    if (isinstance(obj, cfg.ConfigOpts) or
            isinstance(obj, cfg.ConfigOpts.GroupAttr)):
        return {osloconfig_json_serialize(k): osloconfig_json_serialize(v)
                for k, v in obj.items()}
    return obj


def convert_route_targets(orig_list):
    assert isinstance(orig_list, list)
    list_ = []
    for rt in orig_list:
        if rt == '':
            continue
        try:
            asn, nn = rt.split(':')
            list_.append(exa.RouteTarget(int(asn), int(nn)))
        except Exception:
            raise Exception("Malformed route target: '%s'" % rt)
    return list_



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_utils import uuidutils
import testscenarios

from neutron.tests.fullstack.resources import machine

from networking_bagpipe.tests.fullstack import base

load_tests = testscenarios.load_tests_apply_scenarios


class TestBaGPipeML2ConnectivitySameNetwork(base.BaGPipeBaseFullStackTestCase):

    bgpvpn = False
    ipvpn_encap = None
    mech_drivers = 'bagpipe'
    service_plugins = 'router'

    l2_agent_type = constants.AGENT_TYPE_LINUXBRIDGE
    of_interface = None

    compute_node_count = 5
    port_per_compute_per_net = 2

    scenarios = [
        ('BaGPipe native VXLAN', {'bagpipe_ml2': True,
                                  'evpn_driver': 'linux'})]

    def test_connectivity(self):
        tenant_uuid = uuidutils.generate_uuid()

        network = self.safe_client.create_network(tenant_uuid)
        self.safe_client.create_subnet(
            tenant_uuid, network['id'], base.SUBNET_CIDR1)

        vms = machine.FakeFullstackMachinesList([
            self.useFixture(
                machine.FakeFullstackMachine(
                    self.environment.hosts[i],
                    network['id'],
                    tenant_uuid,
                    self.safe_client))
            for i in
            range(self.compute_node_count)*self.port_per_compute_per_net])

        vms.block_until_all_boot()
        vms.ping_all()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from functools import reduce
import threading

from oslo_log import log as logging
import six
from six.moves import queue

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import bgp_peer_worker
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import worker as worker_m


LOG = logging.getLogger(__name__)


if six.PY3:
    def cmp(a, b):
        return (a > b) - (b > a)


class Match(object):

    def __init__(self, afi, safi, route_target):
        assert route_target is None or isinstance(route_target,
                                                  exa.RouteTarget)
        self.afi = afi
        self.safi = safi
        self.route_target = route_target

    def __hash__(self):
        # FIXME, could use a tuple, but RT not yet hashable
        # return hash((self.afi, self.safi, self.route_target))
        return hash(str(self))

    def __repr__(self):
        return "match:%s" % str(self)

    def __str__(self):
        return "%s/%s,%s" % (self.afi or "*", self.safi or "*",
                             self.route_target or "*")

    def __cmp__(self, other):
        assert isinstance(other, Match)

        self_afi = self.afi or engine.Subscription.ANY_AFI
        self_safi = self.safi or engine.Subscription.ANY_SAFI
        self_rt = self.route_target or exa.RouteTarget(0, 0)

        other_afi = other.afi or engine.Subscription.ANY_AFI
        other_safi = other.safi or engine.Subscription.ANY_SAFI
        other_rt = other.route_target or exa.RouteTarget(0, 0)

        val = cmp((self_afi, self_safi, str(self_rt)),
                  (other_afi, other_safi, str(other_rt)))

        return val

    def __lt__(self, other):
        return self.__cmp__(other) == -1

    def __eq__(self, other):
        return self.__cmp__(other) == 0


MATCH_ANY = Match(engine.Subscription.ANY_AFI,
                  engine.Subscription.ANY_SAFI,
                  engine.Subscription.ANY_RT)


def matches_for(afi, safi, route_targets):
    # generate all possible match entries for this afi/safi
    # and these routetargets, with all possible wildcards
    #
    # There are 4*(n+1) possible Match object (for n route_targets)
    for _afi in (engine.Subscription.ANY_AFI, afi):
        for _safi in (engine.Subscription.ANY_SAFI, safi):
            yield Match(_afi, _safi, None)
            if route_targets is not None:
                for rt in route_targets:
                    yield Match(_afi, _safi, rt)


class WorkersAndEntries(object):

    def __init__(self):
        self.workers = set()
        self.entries = set()
        self.n_local_workers = 0

    def __repr__(self):
        return "workers: %s\nentries: %s" % (self.workers,
                                             self.entries)

    def add_worker(self, worker):
        '''returns True iif first local worker'''
        self.workers.add(worker)
        if not isinstance(worker, bgp_peer_worker.BGPPeerWorker):
            self.n_local_workers += 1
            return self.n_local_workers == 1

    def del_worker(self, worker):
        '''returns True iif last local worker'''
        self.workers.remove(worker)
        if not isinstance(worker, bgp_peer_worker.BGPPeerWorker):
            self.n_local_workers -= 1
            return self.n_local_workers == 0

    def is_empty(self):
        return (len(self.workers) == 0 and
                len(self.entries) == 0)


STOP_EVENT = "STOP_EVENT"


def test_should_dispatch(route_event, target_worker):
    '''tests if route should be dispatched to a worker

    returns a (boolean,string) tuple
    the string contains the reason why the route_event should not be
    dispatched to target_worker
    '''
    if route_event.source == target_worker:
        return (False, "not dispatching an update back to its source")
    elif (isinstance(route_event.source, bgp_peer_worker.BGPPeerWorker) and
          isinstance(target_worker, bgp_peer_worker.BGPPeerWorker)):
        return (False, "do not dispatch a route between BGP peers")
    else:
        return (True, "")


class RouteTableManager(threading.Thread, lg.LookingGlassMixin):
    """Singleton class dispatching route events between workers

    Events relates to BGP routes that are announced or withdrawn by workers.
    Workers subscribe to events by indicating AFI, SAFI, Route Targets in which
    they are interested.
    There can be workers responsible for handling services locally (e.g. a VPN
    instance) and workers that are BGP peers.

    Though sources of routes are typically Workers, they don't need to be; any
    class can source a route.
    """

    def __init__(self, first_local_subscriber_cb, last_local_subscriber_cb):
        threading.Thread.__init__(self, name="RouteTableManager")
        self.setDaemon(True)

        # keys are Matches, values are WorkersAndEntries objects:
        self._match_2_workers_entries = (
            collections.defaultdict(WorkersAndEntries))

        # workers known to us
        # name -> worker dict
        self._workers = {}

        # keys are (source,nlri) tuples, values are Entry objects:
        self._source_nlri_2_entry = {}

        self.first_local_subscriber_callback = first_local_subscriber_cb
        self.last_local_subscriber_callback = last_local_subscriber_cb

        self._queue = queue.Queue()

    @log_decorator.log_info
    def stop(self):
        self.enqueue(STOP_EVENT)

    def run(self):
        while True:
            LOG.debug("RouteTableManager waiting on queue")
            event = self._queue.get()
            if event == STOP_EVENT:
                LOG.info("STOP_EVENT => breaking main loop")
                break
            else:
                self._on_event(event)
            LOG.debug("RouteTableManager queue size: %d", self._queue.qsize())

        LOG.info("Out of main loop")

    @log_decorator.log_info
    def _on_event(self, event):
        try:
            if event.__class__ == engine.RouteEvent:
                self._receive_route_event(event)
            elif event.__class__ == engine.Subscription:
                self._worker_subscribes(event)
            elif event.__class__ == engine.Unsubscription:
                self._worker_unsubscribes(event)
            elif event.__class__ == engine.WorkerCleanupEvent:
                self._worker_cleanup(event.worker)
            else:
                raise Exception("unknown event: %s", event)
        except Exception:
            LOG.exception("Exception during processing of event %s", event)

    def enqueue(self, event):
        self._queue.put(event)

    def _check_match_2_workers_and_entries_cleanup(self, match):
        wa = self._match_2_workers_entries.get(match)
        if wa is None:
            LOG.warning("why are we here ?")
            # nothing to cleanup
            return
        else:
            if wa.is_empty():
                del self._match_2_workers_entries[match]

    def callback_first_local_subscriber(self, sub):
        if self.first_local_subscriber_callback:
            LOG.debug("first local subscriber callback for %s ...", sub)
            self.first_local_subscriber_callback(sub)

    def callback_last_local_subscriber(self, sub):
        if self.last_local_subscriber_callback:
            LOG.debug("last local subscriber callback for %s ...", sub)
            self.last_local_subscriber_callback(sub)

    @log_decorator.log_info
    def _worker_subscribes(self, sub):
        assert isinstance(sub.worker, worker_m.Worker)

        worker = sub.worker

        self._workers[worker.name] = worker

        match = Match(sub.afi, sub.safi, sub.route_target)

        wa = self._match_2_workers_entries[match]

        # update match2worker
        if wa.add_worker(worker):
            self.callback_first_local_subscriber(sub)

        LOG.debug("match2workers: %s", wa.workers)

        # re-synthesize events
        for entry in wa.entries:
            LOG.debug("Found an entry for this match: %s", entry)
            event = engine.RouteEvent(engine.RouteEvent.ADVERTISE, entry)
            (dispatch, reason) = test_should_dispatch(event, worker)

            if dispatch:
                # check if the entry carries a route_target to which the worker
                # was already subscribed
                for rt in entry.route_targets:
                    if Match(entry.afi,
                             entry.safi,
                             rt) in worker._rtm_matches:
                        (dispatch, reason) = (
                            False,
                            "worker already had a subscription for this route")
                        break

            if dispatch:
                LOG.info("Dispatching re-synthesized event for %s", entry)
                worker.enqueue(event)
            else:
                LOG.info("%s => not dispatching re-synthesized event for %s",
                         reason, entry)

        # update worker matches
        worker._rtm_matches.add(match)

        # self._dump_state()

    def _worker_unsubscribes(self, sub):
        assert isinstance(sub.worker, worker_m.Worker)

        worker = sub.worker

        # self._dump_state()

        match = Match(sub.afi, sub.safi, sub.route_target)

        # update worker matches
        try:
            worker._rtm_matches.remove(match)
        except KeyError:
            LOG.warning("worker %s unsubs' from %s but this match was"
                        " not tracked for this worker (should not happen,"
                        " this is a bug)", worker, match)

        # synthesize withdraw events
        wa = self._match_2_workers_entries.get(match)
        if wa:
            for entry in wa.entries:
                intersect = set(matches_for(entry.afi,
                                            entry.safi,
                                            entry.route_targets)
                                ).intersection(worker._rtm_matches)
                if len(intersect) > 0:
                    LOG.debug("Will not synthesize withdraw event for %s, "
                              "because worker subscribed to %s", entry,
                              intersect)
                else:
                    LOG.debug("Found an entry for this match: %s", entry)
                    event = engine.RouteEvent(engine.RouteEvent.WITHDRAW,
                                              entry)
                    (dispatch, reason) = test_should_dispatch(event, worker)
                    if dispatch:
                        LOG.info("Dispatching re-synthesized event for %s",
                                 entry)
                        worker.enqueue(event)
                    else:
                        LOG.info("%s => not dispatching re-synthesized event"
                                 " for %s", reason, entry)

            # update _match_2_workers_entries
            try:
                if wa.del_worker(worker):
                    LOG.debug("see if need to callback on last local worker")
                    self.callback_last_local_subscriber(sub)
                    self._check_match_2_workers_and_entries_cleanup(match)
            except KeyError:
                LOG.warning("worker %s unsubscribed from %s but was not"
                            " subscribed yet", worker, match)
        else:  # wa is None
            LOG.warning("worker %s unsubscribed from %s but we had no such"
                        " subscription yet", worker, match)

        if len(worker._rtm_matches) == 0:
            self._workers.pop(worker.name, None)

        # self._dump_state()

    @log_decorator.log
    def _propagate_route_event(self, route_event, except_workers=None):
        '''Propagate route_event to workers subscribed to the route

        based on subscribed RTs/wildcards, and excluding the workers in
        except_workers.

        Returns the list of workers to which the event was propagated.
        '''

        re = route_event.route_entry

        if except_workers is None:
            except_workers = []

        target_workers = set()
        for match in matches_for(re.afi, re.safi, re.route_targets):
            LOG.debug("Finding interested workers for match %s", match)

            interested_workers = self._match_2_workers_entries[match].workers

            LOG.debug("   Workers interested in this match: %s",
                      interested_workers)

            for worker in interested_workers:
                (dispatch, reason) = test_should_dispatch(route_event, worker)
                if dispatch:
                    if worker not in except_workers:
                        LOG.debug("Will dispatch event to %s: %s",
                                  worker, route_event)
                        target_workers.add(worker)
                    else:
                        LOG.debug("Decided not to dispatch to %s, based on"
                                  " except_workers: %s", worker, route_event)
                else:
                    LOG.debug("Decided not to dispatch to %s: %s (%s)",
                              worker, reason, route_event)

        for worker in target_workers:
            LOG.info("Dispatching event to %s: %s", worker, route_event)
            worker.enqueue(route_event)

        return target_workers

    @log_decorator.log_info
    def _receive_route_event(self, route_event):
        entry = route_event.route_entry

        LOG.debug("Try to find an entry from same worker with same nlri")
        replaced_entry = self._source_nlri_2_entry.get((entry.source,
                                                        entry.nlri))

        LOG.debug("   Result: %s", replaced_entry)

        # replaced_entry should be non-empty for a withdraw
        if (replaced_entry is None and
                route_event.type == engine.RouteEvent.WITHDRAW):
            LOG.warning("WITHDRAW but found no route that we could remove: %s",
                        route_event.route_entry)
            return

        # Propagate events to interested workers...
        if route_event.type == engine.RouteEvent.ADVERTISE:

            if replaced_entry == route_event.route_entry:
                LOG.warning("Ignoring, the route advertized is the same as the"
                            " one previously advertized by the source (%s)",
                            route_event.source)
                return

            # propagate event to interested worker
            # and include the info on the route are replaced by this
            # route, if any
            route_event.set_replaced_route(replaced_entry)

            workers_already_notified = self._propagate_route_event(route_event)
        else:  # WITHDRAW
            workers_already_notified = None

        # Synthesize and dispatch a withdraw event for the route entry that
        # was withdrawn or replaced, except, in the case of a replaced route,
        # to workers that had the ADVERTISE event
        if replaced_entry is not None:
            LOG.debug("Synthesizing a withdraw event for replaced route %s",
                      replaced_entry)
            removal_event = engine.RouteEvent(engine.RouteEvent.WITHDRAW,
                                              replaced_entry,
                                              route_event.source)

            self._propagate_route_event(removal_event,
                                        workers_already_notified)

            # Update match2entries for the replaced_route
            for match in matches_for(replaced_entry.afi,
                                     replaced_entry.safi,
                                     replaced_entry.route_targets):
                wa = self._match_2_workers_entries.get(match, None)
                if wa is None:
                    LOG.error("Trying to remove a route from a match, but"
                              " match %s not found - not supposed to happen"
                              " (route: %s)", match, replaced_entry)
                else:
                    wa.entries.discard(replaced_entry)
                    self._check_match_2_workers_and_entries_cleanup(match)

            # update the route entries for this source
            replaced_entry.source._rtm_route_entries.discard(replaced_entry)

        if route_event.type == engine.RouteEvent.ADVERTISE:
            # Update match2entries and source2entries for the newly
            # advertized route
            for match in matches_for(entry.afi,
                                     entry.safi,
                                     entry.route_targets):
                self._match_2_workers_entries[match].entries.add(entry)

            entry.source._rtm_route_entries.add(entry)

            # Update _source_nlri_2_entry
            self._source_nlri_2_entry[(entry.source, entry.nlri)] = entry
        else:  # WITHDRAW
            # Update source2entries
            entry.source._rtm_route_entries.discard(entry)

            # Update _source_nlri_2_entry
            try:
                del self._source_nlri_2_entry[(entry.source, entry.nlri)]
            except KeyError:
                LOG.error("BUG: withdraw, but nothing could be removed in "
                          "_source_nlri_2_entry")

        #  self._dump_state()

    @log_decorator.log_info
    def _worker_cleanup(self, worker):
        '''Cleanup the subscriptions and routes advertised by a worker

        Consider all routes announced by this worker as withdrawn.
        Consider this worker unsubscribed from all of its current
        subscriptions.
        '''
        assert isinstance(worker, worker_m.Worker)
        # synthesize withdraw events for all routes from this worker
        LOG.info("  Preparing to withdraw %d routes that were advertised "
                 "by worker", len(worker._rtm_route_entries))
        for entry in worker._rtm_route_entries:
            LOG.info("  Enqueue event to Withdraw route %s", entry)
            self.enqueue(engine.RouteEvent(engine.RouteEvent.WITHDRAW, entry))

        # remove worker from all of its subscriptions
        for match in worker._rtm_matches:
            wa = self._match_2_workers_entries[match]
            if wa.del_worker(worker):
                self.callback_last_local_subscriber(
                    engine.Subscription(match.afi, match.safi,
                                        match.route_target, worker))
            self._check_match_2_workers_and_entries_cleanup(match)

        worker._rtm_matches.clear()

        # self._dump_state()

    def _dump_state(self):
        if not LOG.isEnabledFor(logging.DEBUG):
            return

        dump = []

        dump.append("~~~ Worker -> Matches ~~~")
        for worker in self._workers.values():
            dump.append("  %s" % worker)
            matches = list(worker._rtm_matches)
            matches.sort()
            for match in matches:
                dump.append("    %s" % match)

        match_2_worker_dump = []
        match_2_entries_dump = []

        matches = list(self._match_2_workers_entries.keys())
        matches.sort()
        for match in matches:
            match_2_worker_dump.append("  %s" % match)
            match_2_entries_dump.append("  %s" % match)
            wa = self._match_2_workers_entries.get(match)
            if wa:
                for worker in wa.workers:
                    match_2_worker_dump.append("    %s" % worker)
                for re in wa.entries:
                    match_2_entries_dump.append("    %s" % re)

        dump.append("\n~~~ Match -> Workers ~~~\n%s\n" %
                    "\n".join(match_2_worker_dump))

        dump.append("~~~ Match -> Entries ~~~\n%s\n" %
                    "\n".join(match_2_entries_dump))

        dump.append("~~~ (source,nlri) ->  entries ~~~")
        for ((source, nlri), entry) in six.iteritems(
                self._source_nlri_2_entry):
            dump.append("  (%s, %s): %s" % (source, nlri, entry))

        LOG.debug("RouteTableManager data dump:\n\n%s\n", "\n".join(dump))

    # Looking Glass #####

    def get_lg_map(self):
        return {"workers": (lg.COLLECTION, (self.get_lg_worker_list,
                                            self.get_lg_worker_from_path_item)
                            ),
                "routes": (lg.SUBTREE, self.get_lg_routes)}

    def get_lg_routes(self, path_prefix):
        result = {}

        match_IPVPN = Match(exa.AFI.ipv4, exa.SAFI.mpls_vpn,
                            engine.Subscription.ANY_RT)
        match_EVPN = Match(exa.AFI.l2vpn, exa.SAFI.evpn,
                           engine.Subscription.ANY_RT)
        match_RTC = Match(exa.AFI.ipv4, exa.SAFI.rtc,
                          engine.Subscription.ANY_RT)
        match_FlowSpecVPN = Match(exa.AFI.ipv4, exa.SAFI.flow_vpn,
                                  engine.Subscription.ANY_RT)
        for match in [match_IPVPN, match_EVPN, match_RTC, match_FlowSpecVPN]:
            match_result = []
            wa = self._match_2_workers_entries.get(match)
            if wa is not None:
                for entry in wa.entries:
                    match_result.append(
                        entry.get_looking_glass_info(path_prefix))
            result[str(match)] = match_result
        return result

    def get_lg_worker_list(self):
        return [{"id": worker.name}
                for worker in six.itervalues(self._workers)]

    def get_lg_worker_from_path_item(self, path_item):
        return self._workers.get(path_item, None)

    def get_all_routes_but_rtc(self):
        return [re for re in self._match_2_workers_entries[MATCH_ANY].entries
                if (re.afi, re.safi) != (exa.AFI.ipv4, exa.SAFI.rtc)
                ]

    def get_local_routes_count(self):
        return reduce(
            lambda count, entry:
            count + (not isinstance(entry.source,
                                    bgp_peer_worker.BGPPeerWorker)),
            self.get_all_routes_but_rtc(),
            0)

    def get_received_routes_count(self):
        return reduce(
            lambda count, entry:
            count + isinstance(entry.source, bgp_peer_worker.BGPPeerWorker),
            self.get_all_routes_but_rtc(),
            0)
